#!/bin/bash
# -*- coding: utf-8 -*-

##
## @(#) リモートサーバ上でタスクスクリプトを実行する
##

set -e           # スクリプトにエラーがあれば終了
set -u           # 未定義の変数があればエラーにする
set -o pipefail  # パイプのコマンドが失敗したらエラー

## エラーメッセージを表示して終了
abort () {
  echo "Error: $1" >&2
  exit 1
}

## コマンドを表示してから実行
run () {
  echo -e "\033[34m[localhost]$ $1\033[m" >&2    # 青色で表示
  eval $@
}

## ヘルプメッセージを表示
print_help () {
  local roles=`echo *.role | sed 's/\.role//g; s/ /, /g;'`
  [[ "$roles" = "*" ]] && roles='(none)'
  echo "Usage: `basename $0` <role>"
  echo "Roles: $roles"
}

## ・引数がなければ、ヘルプを表示して終了
## ・引数が -h または --help でも、ヘルプを表示して終了
## ・引数が1個だけならロール名、それより多ければエラー
[[ $# -eq 0      ]] && { print_help; exit 0; }
[[ $1 = '-h'     ]] && { print_help; exit 0; }
[[ $1 = '--help' ]] && { print_help; exit 0; }
[[ $# -gt 1      ]] && { abort "too many arguments."; }
role=$1
[[ -d $role.role ]] || abort "$role.role: directory not exist."

## 環境変数 $APP_ENV が設定されてなければエラー
[[ -n "${APP_ENV:-}" ]] || abort '$APP_ENV should be set.'

## ロールディレクトリに移動
cd $role.role

## sshの設定
. hosts/ssh_${APP_ENV}.bash  # 設定ファイルを読み込む
: $ssh_user      # 変数が設定されていることを確認（されてないとエラー）
: $ssh_port
: $ssh_key
: $ssh_option

## sshオプションを追加
[[ -z "$ssh_key"  ]] || ssh_option="-i $ssh_key $ssh_option"
[[ -z "$ssh_port" ]] || ssh_option="-p $ssh_port $ssh_option"

## rsyncのオプション
##   -a: まるごとコピー、-v: 送信ファイル名を出力、-z: 圧縮して通信、
##   -L: シンボリックリンクならリンク先のファイルをコピー、
##   --delete: リストにないファイルは削除、-e: 接続に使うsshコマンド
rsync_option="-avzL --delete -e 'ssh $ssh_option'"

## リモートサーバにコピーするファイル
files="tasks.txt tasks files vars lib ../lib/remote.sh"

## 環境別の変数ファイルが存在することを確認
varfile="vars/vars_${APP_ENV}.bash"
[[ -f $varfile ]] || abort "$varfile: file not exist."

## サーバ名をホストファイルから読み出す
## （注：while文だとsshの -t オプションがエラーになるので、whileを避ける）
hostsfile="hosts/hosts_${APP_ENV}.txt"
for ssh_host in `grep -Ev '^#|^[ \t]*$' $hostsfile`; do

  ## rsyncでリモートサーバにファイルをコピー
  remote=${ssh_user}@${ssh_host}  # 例：ubuntu@localhost
  run "rsync $rsync_option $files $remote:$role.role/"

  ## sshでリモートサーバに接続し、スクリプトを実行
  cmd="APP_ENV=$APP_ENV sh remote.sh $remote $role"
  run "ssh -t $ssh_option $remote 'cd $role.role; $cmd'"

done
