#!/bin/bash
# -*- coding: utf-8 -*-

##
## @(#) リモートサーバ上でタスクスクリプトを実行する
##

set -e           # スクリプトにエラーがあれば終了
set -u           # 未定義の変数があればエラーにする
set -o pipefail  # パイプのコマンドが失敗したらエラー

## エラーメッセージを表示して終了
abort () {
  echo "Error: $1" >&2
  exit 1
}

## コマンドを表示してから実行
run () { _pr $@; eval $@; }
_pr () { echo -e "\033[34m[localhost]$ $@\033[m" >&2; } # 青色で表示

## ヘルプメッセージを表示
print_help () {
  local roles=`echo *.role | sed 's/\.role//g; s/ /, /g;'`
  [[ "$roles" = "*" ]] && roles='(none)'
  echo "Usage: `basename $0` <role>"
  echo "Roles: $roles"
}

## ・引数がなければ、ヘルプを表示して終了
## ・引数が -h または --help でも、ヘルプを表示して終了
## ・引数が1個だけならロール名、それより多ければエラー
[[ $# -eq 0      ]] && { print_help; exit 0; }
[[ $1 = '-h'     ]] && { print_help; exit 0; }
[[ $1 = '--help' ]] && { print_help; exit 0; }
[[ $# -gt 1      ]] && { abort "too many arguments."; }
role=$1
[[ -d $role.role ]] || abort "$role.role: directory not exist."

## 環境変数 $APP_ENV が設定されてなければエラー
[[ -n "${APP_ENV:-}" ]] || abort '$APP_ENV should be set.'

## ロールディレクトリに移動
cd $role.role

## sshの設定
. hosts/ssh_${APP_ENV}.bash  # 設定ファイルを読み込む
: $ssh_user      # 変数が設定されていることを確認（されてないとエラー）
: $ssh_port
: $ssh_key
: $ssh_option

## sshオプションを追加
[[ -z "$ssh_key"  ]] || ssh_option="-i $ssh_key $ssh_option"
[[ -z "$ssh_port" ]] || ssh_option="-p $ssh_port $ssh_option"

## rsyncのオプション
##   -a: まるごとコピー、-v: 送信ファイル名を出力、-z: 圧縮して通信、
##   -L: シンボリックリンクならリンク先のファイルをコピー、
##   --delete: リストにないファイルは削除、-e: 接続に使うsshコマンド
rsync_option="-avzL --delete -e 'ssh $ssh_option'"

## リモートサーバにコピーするファイル
files="tasks.txt tasks files vars lib ../lib/remote.sh"

## ロール全体の前処理（on ローカルマシン）
local="lib/local.bash"
bash $local before-role $role

## パスワードファイルが置かれているディレクトリを検証する関数
check_password_dir () {
  local dir=$1
  ## ディレクトリがなければエラー
  [ -d "$dir" ] || abort "$dir: directory not exist"
  ## シンボリックリンクならセキュリティの懸念があるためエラー
  [ -L "$dir" ] && abort "$dir: should not be symbolic link."
  ## 所有者が root ユーザでなければエラー
  set -- `sudo ls -ld $dir`
  local permission=$1 owner=$3
  [ "$owner" = "root" ] || abort "$dir: should be owned by root."
  ## 所有者である root ユーザ以外がアクセスできるならエラー
  case "$permission" in
  d*------) ;;   # ok
  *)  abort "$dir: non-root user can access to directory." ;;
  esac
}

## パスワードファイルを読み込む関数
read_password_file () {
  local pwfile pw
  ## 環境変数 $PASSWORD_FILE にパス名が設定されていること
  pwfile=${PASSWORD_FILE:-}
  [ -n "$pwfile" ] || abort "\$PASSWORD_FILE should be set."
  ## パスワードファイルを読み込む
  pw=`run 'sudo -p "(sudo) Password:" cat $PASSWORD_FILE'`
  ## パスワードファイルが置かれているディレクトリを検証
  check_password_dir `dirname $PASSWORD_FILE`
  ## チェックに通ったらパスワードを返す
  echo $pw
}

## 暗号化されたファイルを残して、復号されたファイルを消す関数
## （暗号化されたファイルがあれば、もとのファイルはいらないはず）
remove_decrypted_files () {
  local encrypted plain msg answer
  { find . -name '*.encrypted' | while read -r encrypted; do
      plain=${encrypted%.encrypted}  # 拡張子 .encrypted を変数展開で取り除く
      [[ -f "$plain" ]] || continue  # 復号されたファイルがなければスキップ
      msg="${plain}: remove decrypted file? [y/N] "
      read -rp "$msg" answer <&3     # 端末からユーザ入力を読み込む
      case "$answer" in
      y*|Y*) run "rm $plain";;       # ユーザ入力が y か Y で始まるなら、削除
      esac
    done
  } 3<&0                             # 標準入力（端末）を別の入力に結びつける
}

## 暗号化されたファイルがあれば、パスワードファイルを読み込む
password=''
n=`find . -name '*.encrypted' | wc -l`
[[ $n -eq 0 ]] || password=`read_password_file`
[[ $n -eq 0 ]] || remove_decrypted_files  # もとのファイルを消す

## リモートサーバが再起動されるまで、再接続を繰り返す
wait_until_reboot () {
  remote="$1"; ssh_option="$2"
  echo -n "Re-connecting "
  sleep 5    # リモートサーバのsshdが終了するのを待つ
  echo -n "."
  ## 再接続がタイムアウト（3秒）したら、1秒待ってやり直す
  while ! ssh $ssh_option -q -o 'ConnectTimeout 3' $remote ":"; do
    sleep 1
    echo -n "."
  done
  echo
  echo "Connected!"
}

## サーバ名をホストファイルから読み出す
## （注：while文だとsshの -t オプションがエラーになるので、whileを避ける）
hostsfile="hosts/hosts_${APP_ENV}.txt"
for ssh_host in `grep -Ev '^#|^[ \t]*$' $hostsfile`; do

  ## ホストごとの前処理（on ローカルマシン）
  bash $local before-host $role $ssh_host

  ## rsyncでリモートサーバにファイルをコピー
  remote=${ssh_user}@${ssh_host}  # 例：ubuntu@localhost
  run rsync $rsync_option $files $remote:$role.role/

  ## sshでリモートサーバに接続し、スクリプトを実行
  set +e      # 終了ステータスを調べる必要があるので、
              # 終了ステータスが0でなくてもスクリプトを終了しない
  REBOOT_EXIT=246    # 再接続を表す終了ステータス
  cmd="APP_ENV=$APP_ENV sh remote.sh $remote $role"
  _pr "ssh -t $ssh_option $remote 'cd $role.role; $cmd'"
  ssh      -t $ssh_option $remote "cd $role.role; $cmd $password"
  status=$?          # sshの終了ステータス
  while [[ $status -eq $REBOOT_EXIT ]]; do   # 再接続が必要なら
    wait_until_reboot $remote "$ssh_option" >&2  # 再起動を待つ
    _pr "ssh -t $ssh_option $remote 'cd $role.role; $cmd'"
    ssh      -t $ssh_option $remote "cd $role.role; $cmd $password"
    status=$?
  done
  set -e      # もとに戻す

  ## ホストごとの後処理（on ローカルマシン）
  bash $local after-host $role $ssh_host

  ## サーバ上でスクリプトの実行にエラーがあれば、途中で終了
  if [[ $status -ne 0 ]]; then
    { echo
      echo "***"
      echo "*** ERROR: task failed on $remote (status=$status)."
      echo "***"
    } >&2
    exit $status
  fi

done

## ロール全体の後処理（on ローカルマシン）
bash $local after-role $role
